#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <string.h>

#include "common.h"

/*
** The environment variable that the sample applications
** attempt to pick up and use to find the base of the ADM-XRC
** SDK and hence the FPGA bitstream files.
*/
#define ENVIRONMENT_VARIABLE ("ADMXRC_SDK4")

#ifdef WIN32
static const char nativeSlash = '\\';
static const char nativeSlashStr[] = "\\";
static const char nonNativeSlash = '/';
static const char nonNativeSlashStr[] = "/";
#else
static const char nativeSlash = '/';
static const char nativeSlashStr[] = "/";
static const char nonNativeSlash = '\\';
static const char nonNativeSlashStr[] = "\\";
#endif

static const char*	typeString[] = {
	"float", "float", "int", "uint",
	"hex", "string", "boolean", "path"
};

#define SET_STRING(STRING, VALUE) \
{ \
	if ((STRING) != NULL) \
		free(STRING); \
	if ((VALUE) != NULL) \
	{ \
		(STRING) = malloc(sizeof(char) * (strlen(VALUE) + 1)); \
		(STRING)[0] = '\0'; \
		strcpy((STRING), (VALUE)); \
	} \
	else \
		(STRING) = NULL; \
}

static void fatal(int code, char* fmt, ...)
{
	char		buffer[1024];
	va_list		ap;
	size_t		n;

	va_start(ap, fmt);
	n = _vsnprintf(buffer, 1023, fmt, ap);
	va_end(ap);
	buffer[n] = '\0';

	printf(buffer);

	exit(code);
}

static void convert_slashes(char* str)
{
	while (*str != '\0')
	{
		if (*str == nonNativeSlash)
			*str = nativeSlash;
		str++;
	}
}

static void strip_leading_space(char* str)
{
	char* p;
	char* q;

	p = str;
	q = str;

	while (*p != '\0' && isspace(*p))
		p++;
	while (*p != '\0')
		*q++ = *p++;
	*q = 0;
}

static void strip_trailing_space(char* str)
{
	int		i, n;

	n = strlen(str);
	if (n == 0) {
		return;
	}

	for (i = n - 1; i >= 0; i--) {
		if (!isspace(str[i])) {
			i++;
			break;
		}
	}

	str[i] = '\0';
}

static void strip_trailing_slash(char* str)
{
	int		n;

	n = strlen(str);
	if (n) {
		if (str[n - 1] == nativeSlash) {
			if (n != 1) {
				str[n - 1] = '\0';
			}
		}
	}
}

static void tidy_path(char* str)
{
	convert_slashes(str);
	strip_leading_space(str);
	strip_trailing_space(str);
	strip_trailing_slash(str);
}

int set_value(
	Value*		value,
	OptionType	type,
	char*		string)
{
	char buff[5];
	union {
		float			floatVal;
		long			intVal;
		unsigned long	uintVal;
		BOOLEAN			booleanVal;
	} val;

	if (type == Option_float) {
		sscanf(string, "%f", &val.floatVal);
		value->floatVal = val.floatVal;
	} else if (type == Option_double) {
		sscanf(string, "%f", &val.floatVal);
		value->doubleVal = (double) val.floatVal;
	} else if (type == Option_int) {
		sscanf(string, "%ld", &val.intVal);
		value->intVal = val.intVal;
	} else if (type == Option_uint) {
		if (strlen(string) >= 2 && string[0] == '0' && tolower(string[1]) == 'x') {
			sscanf(string + 2, "%lx", &val.uintVal);
		} else {
			sscanf(string, "%lu", &val.uintVal);
		}
		value->uintVal = val.uintVal;
	} else if (type == Option_hex) {
		sscanf(string, "%lx", &val.uintVal);
		value->hexVal = val.uintVal;
	} else if (type == Option_string) {
		SET_STRING(value->stringVal, string);
	} else if (type == Option_boolean) {
		strncpy(buff, string, 5);
		buff[4] = '\0';
		strlwr(buff);
		if (strcmp(buff, "true") == 0) {
			value->booleanVal = TRUE;
		} else if (strcmp(buff, "true") == 0) {
			value->booleanVal = FALSE;
		} else {
			return -1;
		}
	} else if (type == Option_path) {
		SET_STRING(value->pathVal, string);
		tidy_path(value->pathVal);
	} else {
		printf("*** internal error in arg.c:set_value()\n");
		return -1;
	}

	return 0;
}

static char*
default_to_string(
	Option*	option)
{
	static char buffer[128];

	switch(option->type)
	{
	case Option_float:
		sprintf(buffer, "%f", (float) option->def.floatVal); break;
	case Option_double:
		sprintf(buffer, "%f", (float) option->def.doubleVal); break;
	case Option_int:
		sprintf(buffer, "%ld", option->def.intVal); break;
	case Option_uint:
		sprintf(buffer, "%lu", option->def.uintVal); break;
	case Option_hex:
		sprintf(buffer, "0x%lx", option->def.hexVal); break;
	case Option_string:
		if (option->def.stringVal != NULL) {
			sprintf(buffer, "%s", option->def.stringVal);
		} else {
			strcpy(buffer, NULL);
		}
		break;
	case Option_boolean:
		if (option->def.booleanVal) {
			strcpy(buffer, "TRUE");
		} else {
			strcpy(buffer, "FALSE");
		}
		break;
	case Option_path:
		if (option->def.pathVal != NULL) {
			sprintf(buffer, "%s", option->def.pathVal);
		} else {
			strcpy(buffer, NULL);
		}
		break;
	}

	return buffer;
}

static char*
value_to_string(
	Value*		value,
	OptionType	type)
{
	static char buffer[128];

	switch(type)
	{
	case Option_float:
		sprintf(buffer, "%f", (float) value->floatVal); break;
	case Option_double:
		sprintf(buffer, "%f", (float) value->doubleVal); break;
	case Option_int:
		sprintf(buffer, "%ld", value->intVal); break;
	case Option_uint:
		sprintf(buffer, "%lu", value->uintVal); break;
	case Option_hex:
		sprintf(buffer, "0x%lx", value->hexVal); break;
	case Option_string:
		if (value->stringVal != NULL) {
			sprintf(buffer, "%s", value->stringVal);
		} else {
			strcpy(buffer, NULL);
		}
		break;
	case Option_boolean:
		if (value->booleanVal) {
			strcpy(buffer, "TRUE");
		} else {
			strcpy(buffer, "FALSE");
		}
		break;
	case Option_path:
		if (value->pathVal != NULL) {
			sprintf(buffer, "%s", value->pathVal);
		} else {
			strcpy(buffer, NULL);
		}
		break;
	}

	return buffer;
}

static char*
option_to_string(
	Option*	option)
{
	static char*	val;
	static char		buffer[128];

	val = value_to_string(&option->value, option->type);
	if (option->type == Option_boolean) {
		sprintf(
			buffer,
			"%c%s %s\n",
			option->value.booleanVal ? "+" : "-",
			option->key,
			val);
	} else {
		sprintf(buffer, "-%s %s", option->key, val);
	}

	return buffer;
}

static void set_option_default(
	Option*	option)
{
	switch (option->type)
	{
	case Option_float:
		option->value.floatVal = option->def.floatVal;
		break;
	case Option_double:
		option->value.doubleVal = option->def.doubleVal;
		break;
	case Option_int:
		option->value.intVal = option->def.intVal;
		break;
	case Option_uint:
		option->value.uintVal = option->def.uintVal;
		break;
	case Option_hex:
		option->value.hexVal = option->def.hexVal;
		break;
	case Option_string:
		option->value.stringVal = NULL;
		SET_STRING(option->value.stringVal, option->def.stringVal);
		break;
	case Option_boolean:
		option->value.booleanVal = option->def.booleanVal;
		break;
	case Option_path:
		option->value.pathVal = NULL;
		SET_STRING(option->value.pathVal, option->def.pathVal);
		break;
	}
	option->specified = FALSE;
}

static void set_option(
	Option*		option,
	char*		string)
{
	union {
		float			floatVal;
		long			intVal;
		unsigned long	uintVal;
		BOOLEAN			booleanVal;
	} value;

	if (option->type == Option_float) {
		sscanf(string, "%f", &value.floatVal);
		option->value.floatVal = value.floatVal;
	} else if (option->type == Option_double) {
		sscanf(string, "%f", &value.floatVal);
		option->value.doubleVal = (double) value.floatVal;
	} else if (option->type == Option_int) {
		sscanf(string, "%ld", &value.intVal);
		option->value.intVal = value.intVal;
	} else if (option->type == Option_uint) {
		if (strlen(string) >= 2 && string[0] == '0' && string[1] == 'x') {
			sscanf(string + 2, "%lx", &value.intVal);
		} else {
			sscanf(string, "%lu", &value.intVal);
		}
		option->value.uintVal = value.uintVal;
	} else if (option->type == Option_hex) {
		sscanf(string, "%lx", &value.uintVal);
		option->value.hexVal = value.uintVal;
	} else if (option->type == Option_string) {
		SET_STRING(option->value.stringVal, string);
	} else if (option->type == Option_path) {
		SET_STRING(option->value.pathVal, string);
		tidy_path(option->value.pathVal);
	}
	option->specified = TRUE;
}

static Option* find_option(
	char*		key,
	Option*		options,
	int			n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (strcmp(options[i].key, key) == 0) {
			return &options[i];
		}
	}

	return NULL;
}

void show_usage(
	char*		appName,
	Arguments*	args)
{
	int i;

	printf("Usage: %s [option ...]", appName);
	for (i = 0; i < args->nParamType; i++) {
		char* a, * b;

		if (i < args->minNParam) {
			a = "";
			b = "";
		} else {
			a = "[";
			b = "]";
		}
		printf(" %s%s%s", a, args->paramInfo[i], b);
	}
	if (args->nParamType) {
		printf("\n");
	}
	printf("\n");
	for (i = 0; i < args->nParamType; i++) {
		printf("%10s = %s\n", args->paramInfo[i], args->paramMeaning[i]);
	}
	printf("\n");
	printf("%12s  %7s  %12s  %s\n\n", "Option", "Type", "Default", "Meaning");
	printf("%12s  %7s  %12s  %s\n", "/h /? /help", "", "", "Show this help");
	printf("%12s  %7s  %12s  %s\n", "/sdk", "string", "", "Specify base directory of ADM-XRC SDK");
	printf("%12s  %7s  %12s  %s\n", "/bitdir", "string", "", "Specify bitstream directory");
	for (i = 0; i < args->nOption; i++) {
		char buff[17];

		if (args->option[i].type == Option_boolean) {
			strcpy(buff, "-/+");
		} else {
			strcpy(buff, "/");
		}
		strncat(buff, args->option[i].key, 16);
        buff[16] = '\0';
		printf("%12s  %7s  %12s  %s\n",
			buff,
			typeString[args->option[i].type],
			default_to_string(&args->option[i]),
			args->option[i].meaning);
	}
}

void dump_command_line(
	Arguments*	args)
{
	char*	buffer;
	int		i, j;

	printf("Command line: %s", args->appName);
	for (i = 0; i < args->nOption; i++) {
		if (args->option[i].specified) {
			buffer = option_to_string(&args->option[i]);
			printf(" %s", buffer);
		}
	}
	j = args->nParam;
	if (j > args->nParamType) {
		j = args->nParamType;
	}
	for (i = 0; i < j; i++) {
		buffer = value_to_string(&args->param[i], args->paramType[i]);
		printf(" %s", buffer);
	}
	for (; i < args->nParam; i++) {
		buffer = value_to_string(&args->param[i], Option_string);
		printf(" %s", buffer);
	}
	printf("\n");
}

void
dump_options(
	Arguments*	args)
{
	int i;

	printf("Dump of all options:\n");
	for (i = 0; i < args->nOption; i++) {
		char*				buffer;

		buffer = value_to_string(
			&args->option[i].value,
			args->option[i].type);
		printf(
			"%12s %8s %8s\n",
			args->option[i].key,
			typeString[args->option[i].type],
			buffer);
	}
}

int
parse_command_line(
	char*				appName,
	int					argc,
	char*				argv[],
	Arguments*			args)
{
	char*	admxrc_sdk;
	int		i, j;

	if (appName == NULL) {
		fatal(-1, "parse_command_line(): argument 'appName' is NULL (must not be NULL)\n");
	}
	if (args == NULL) {
		fatal(-1, "parse_command_line(): argument 'args' is NULL (must not be NULL)\n");
	}

	args->appName = NULL;
	args->sdkPath = NULL;
	args->bitPath = NULL;
	args->bitFile = NULL;
	SET_STRING(args->appName, appName);

	for (i = 0; i < args->nOption; i++) {
		set_option_default(&args->option[i]);
	}

	admxrc_sdk = getenv(ENVIRONMENT_VARIABLE);
	if (admxrc_sdk != NULL) {
		SET_STRING(args->sdkPath, admxrc_sdk);
		tidy_path(args->sdkPath);
	}

	// Parse the options
	for (i = 1; i < argc; i++) {
		char tag = argv[i][0];

		if ((tag == '-') || (tag == '/') || (tag == '+')) {
			if ((strcmp(argv[i] + 1, "h") == 0) ||
				(strcmp(argv[i] + 1, "help") == 0) ||
				(strcmp(argv[i] + 1, "?") == 0)) {
				show_usage(appName, args);
				exit(0);
			} else if (strcmp(argv[i] + 1, "sdk") == 0) {
				i++;
				if (i == argc) {
					fatal(-1, "/sdk option requires parameter: path to ADM-XRC SDK root\n");
				}
				SET_STRING(args->sdkPath, argv[i]);
				tidy_path(args->sdkPath);
			} else if (strcmp(argv[i] + 1, "bitdir") == 0) {
				i++;
				if (i == argc) {
					fatal(-1, "/bitdir option requires parameter: path to bitstream directory\n");
				}
				SET_STRING(args->bitPath, argv[i]);
				tidy_path(args->bitPath);
			} else {
				int		narg;
				Option*	option;

				option = find_option(argv[i] + 1, args->option, args->nOption);
				if (option == NULL) {
					fatal(-1, "Unknown option: %s\n", argv[i]);
				} else {
					narg = (option->type == Option_boolean) ? 0 : 1;
					if (i + narg + 1 > argc) {
						fatal(-1, "%s option requires parameter: %s\n",
							argv[i], option->meaning);
					}
					if (option->type == Option_boolean) {
						if (tag == '-') {
							option->value.booleanVal = FALSE;
						} else if (tag == '+') {
							option->value.booleanVal = TRUE;
						} else {
							fatal(-1, "Boolean option %s should be specified as -%s or +%s\n",
								option->key, option->key, option->key);
						}
					} else {
						i++;
						set_option(option, argv[i]);
					}
				}
			}
		}
		else
			break;
	}

	args->nParam = argc - i;
	if (args->nParam == 0) {
		args->param = NULL;
	} else {
		args->param = malloc(sizeof(Value) * args->nParam);
		j = 0;
		for (; i < argc; i++) {
			if (j < args->nParamType) {
				if (args->paramType[j] == Option_string) {
					args->param[j].stringVal = NULL;
				} else if (args->paramType[j] == Option_path) {
					args->param[j].pathVal = NULL;
				}

				if (set_value(&args->param[j], args->paramType[j], argv[i])) {
					fatal(
						-1,
						"Value of %s parameter %d is invalid: %s\n",
						typeString[args->paramType[i]],
						i + 1,
						argv[i]);
				}
			} else {
				set_value(&args->param[j], Option_string, argv[i]);
			}
			j++;
		}
	}

	if (args->minNParam > args->nParam) {
		printf("Too few parameters supplied. Use %s /? to get help.\n", args->appName);
		exit(-1);
	}

	if (args->bitPath == NULL && args->sdkPath != NULL) {
		const char*	bitDir = "bit";
		char* bitpath;
		
		bitpath = (char*) malloc(strlen(args->sdkPath) + 1 + strlen(bitDir) + 1);
		if (bitpath == NULL) {
			fatal(-1, "parse_command_line(): failed to allocate 'bitpath'\n");
		}
		strcpy(bitpath, args->sdkPath);
		strcat(bitpath, nativeSlashStr);
		strcat(bitpath, bitDir);
		SET_STRING(args->bitPath, bitpath);
		free(bitpath);
	}

	return 0;
}

const char*
get_fpga_filename(
	ADMXRC2_CARD_INFO*	cardInfo,
	char*				design,
	Arguments*			args)
{
	return get_fpga_filename_dir(cardInfo, design, design, args);
}

const char*
get_fpga_filename_dir(
	ADMXRC2_CARD_INFO*	cardInfo,
	char*				dir,
	char*				design,
	Arguments*			args)
{
	const char ext[] = ".bit";
	char* board;
	char* device;
	char* bitfile;

	if (design == NULL) {
		fatal(-1, "get_fpga_filename(): argument 'design' is NULL (must not be NULL)\n");
	}

	switch(cardInfo->BoardType)
	{
	case ADMXRC2_BOARD_ADMXRC:
		board = "xrc";
		break;
	case ADMXRC2_BOARD_ADMXRC_P:
		board = "xrcp";
		break;
	case ADMXRC2_BOARD_ADMXRC2_LITE:
		board = "xrc2l";
		break;
	case ADMXRC2_BOARD_ADMXRC2:
		board = "xrc2";
		break;
	case ADMXRC2_BOARD_ADMXPL:
		board = "xpl";
		break;
	case ADMXRC2_BOARD_ADMXP:
		board = "xp";
		break;
	case ADMXRC2_BOARD_ADPWRC2:
		board = "wrc2";
		break;
	case ADMXRC2_BOARD_ADPDRC2:
		board = "drc2";
		break;
	case ADMXRC2_BOARD_ADPXPI:
		board = "xpi";
		break;
	case ADMXRC2_BOARD_ADMXRC4LS:
		board = "xrc4ls";
		break;
	case ADMXRC2_BOARD_ADMXRC4LX:
		board = "xrc4lx";
		break;
	case ADMXRC2_BOARD_ADMXRC4SX:
		board = "xrc4sx";
		break;
	case ADMXRC2_BOARD_ADMXRC4FX:
		board = "xrc4fx";
		break;
	case ADMXRC2_BOARD_ADPEXRC4FX:
		board = "xrce4fx";
		break;
	case ADMXRC2_BOARD_ADMXRC5LX:
		board = "xrc5lx";
		break;
	case ADMXRC2_BOARD_ADMXRC5T1:
		board = "xrc5t1";
		break;
	case ADMXRC2_BOARD_ADMXRC5T2:
		board = "xrc5t2";
		break;
    case ADMXRC2_BOARD_ADCPXRC4LX:
        board = "xrcp4lx";
        break;
	case ADMXRC2_BOARD_ADMAMC5A2:
		board = "amc5a2";
		break;
	case ADMXRC2_BOARD_ADMXRC5TZ:
		board = "xrc5tz";
		break;
	default:
		fatal(-1, "Unknown board type: %d\n",	cardInfo->BoardType);
	}

	switch(cardInfo->FPGAType)
	{
	case ADMXRC2_FPGA_V405E:
		device = "v405e";
		break;
	case ADMXRC2_FPGA_V812E:
		device = "v812e";
		break;
	case ADMXRC2_FPGA_V3200E:
		device = "v3200e";
		break;
	case ADMXRC2_FPGA_V2000E:
		device = "v2000e";
		break;
	case ADMXRC2_FPGA_V1600E:
		device = "v1600e";
		break;
	case ADMXRC2_FPGA_V1000E:
		device = "v1000e";
		break;
	case ADMXRC2_FPGA_V1000:
		device = "v1000";
		break;
	case ADMXRC2_FPGA_V800:
		device = "v800";
		break;
	case ADMXRC2_FPGA_V600:
		device = "v600";
		break;
	case ADMXRC2_FPGA_V400:
		device = "v400";
		break;
	case ADMXRC2_FPGA_2V1000:
		device = "2v1000";
		break;
	case ADMXRC2_FPGA_2V1500:
		device = "2v1500";
		break;
	case ADMXRC2_FPGA_2V2000:
		device = "2v2000";
		break;
	case ADMXRC2_FPGA_2V3000:
		device = "2v3000";
		break;
	case ADMXRC2_FPGA_2V4000:
		device = "2v4000";
		break;
	case ADMXRC2_FPGA_2V6000:
		device = "2v6000";
		break;
	case ADMXRC2_FPGA_2V8000:
		device = "2v8000";
		break;
	case ADMXRC2_FPGA_2V10000:
		device = "2v10000";
		break;
	case ADMXRC2_FPGA_2VP2:
		device = "2vp2";
		break;
	case ADMXRC2_FPGA_2VP4:
		device = "2vp4";
		break;
	case ADMXRC2_FPGA_2VP7:
		device = "2vp7";
		break;
	case ADMXRC2_FPGA_2VP20:
		device = "2vp20";
		break;
	case ADMXRC2_FPGA_2VP30:
		device = "2vp30";
		break;
	case ADMXRC2_FPGA_2VP40:
		device = "2vp40";
		break;
	case ADMXRC2_FPGA_2VP50:
		device = "2vp50";
		break;
	case ADMXRC2_FPGA_2VP100:
		device = "2vp100";
		break;
	case ADMXRC2_FPGA_2VP125:
		device = "2vp125";
		break;
	case ADMXRC2_FPGA_2VP70:
		device = "2vp70";
		break;
	case ADMXRC2_FPGA_4VLX15:
		device = "4vlx15";
		break;
	case ADMXRC2_FPGA_4VLX25:
		device = "4vlx25";
		break;
	case ADMXRC2_FPGA_4VLX40:
		device = "4vlx40";
		break;
	case ADMXRC2_FPGA_4VLX60:
		device = "4vlx60";
		break;
	case ADMXRC2_FPGA_4VLX80:
		device = "4vlx80";
		break;
	case ADMXRC2_FPGA_4VLX100:
		device = "4vlx100";
		break;
	case ADMXRC2_FPGA_4VLX160:
		device = "4vlx160";
		break;
	case ADMXRC2_FPGA_4VLX200:
		device = "4vlx200";
		break;
	case ADMXRC2_FPGA_4VSX25:
		device = "4vsx25";
		break;
	case ADMXRC2_FPGA_4VSX35:
		device = "4vsx35";
		break;
	case ADMXRC2_FPGA_4VSX55:
		device = "4vsx55";
		break;
	case ADMXRC2_FPGA_4VFX12:
		device = "4vfx12";
		break;
	case ADMXRC2_FPGA_4VFX20:
		device = "4vfx20";
		break;
	case ADMXRC2_FPGA_4VFX40:
		device = "4vfx40";
		break;
	case ADMXRC2_FPGA_4VFX60:
		device = "4vfx60";
		break;
	case ADMXRC2_FPGA_4VFX100:
		device = "4vfx100";
		break;
	case ADMXRC2_FPGA_4VFX140:
		device = "4vfx140";
		break;
	case ADMXRC2_FPGA_5VLX30:
		device = "5vlx30";
		break;
	case ADMXRC2_FPGA_5VLX50:
		device = "5vlx50";
		break;
	case ADMXRC2_FPGA_5VLX85:
		device = "5vlx85";
		break;
	case ADMXRC2_FPGA_5VLX110:
		device = "5vlx110";
		break;
	case ADMXRC2_FPGA_5VLX220:
		device = "5vlx220";
		break;
	case ADMXRC2_FPGA_5VLX330:
		device = "5vlx330";
		break;
	case ADMXRC2_FPGA_5VLX30T:
		device = "5vlx30t";
		break;
	case ADMXRC2_FPGA_5VLX50T:
		device = "5vlx50t";
		break;
	case ADMXRC2_FPGA_5VLX85T:
		device = "5vlx85t";
		break;
	case ADMXRC2_FPGA_5VLX110T:
		device = "5vlx110t";
		break;
	case ADMXRC2_FPGA_5VLX330T:
		device = "5vlx330t";
		break;
	case ADMXRC2_FPGA_5VLX220T:
		device = "5vlx220t";
		break;
	case ADMXRC2_FPGA_5VSX35T:
		device = "5vsx35t";
		break;
	case ADMXRC2_FPGA_5VSX50T:
		device = "5vsx50t";
		break;
	case ADMXRC2_FPGA_5VSX95T:
		device = "5vsx95t";
		break;
	case ADMXRC2_FPGA_5VSX240T:
		device = "5vsx240t";
		break;
	case ADMXRC2_FPGA_5VFX100T:
		device = "5vfx100t";
		break;
	case ADMXRC2_FPGA_5VFX130T:
		device = "5vfx130t";
		break;
	case ADMXRC2_FPGA_5VFX200T:
		device = "5vfx200t";
		break;
	default:
		fatal(-1, "Unknown FPGA type on card: %d\n",
			cardInfo->FPGAType);
	}

	if (args->bitPath != NULL) {
		bitfile = malloc(
			strlen(args->bitPath) + 1 +
			strlen(dir) + 1 +
			strlen(design) + 1 +
			strlen(board) + 1 +
			strlen(device) +
			strlen(ext) + 1);
		if (bitfile == NULL) {
			fatal(-1, "get_fpga_filename(): failed to allocate 'bitfile'\n");
		}
		strcpy(bitfile, args->bitPath);
		strcat(bitfile, nativeSlashStr);
		strcat(bitfile, dir);
		strcat(bitfile, nativeSlashStr);
		strcat(bitfile, design);
		strcat(bitfile, "-");
		strcat(bitfile, board);
		strcat(bitfile, "-");
		strcat(bitfile, device);
		strcat(bitfile, ext);
		SET_STRING(args->bitFile, bitfile);
		free(bitfile);
	} else {
		fatal(
			-1,
			"The path to the FPGA bitstream directory cannot be determined via the "
			"ADMXRC_SDK4 environment variable and has not been specified via the "
			"via the /sdk or /bitdir switches. Please correct this and "
			"try again.\n");
	}

	return args->bitFile;
}
