/*
** dma.c - example demonstrating multithreaded DMA transfers
**         to and from FPGA, using demand mode DMA over PCIe
**
** Data is transferred using demand mode DMA to the FPGA, stored
** in local memory and transferred out again also using demand mode DMA.
**
** Example for:
**
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2
**   o ADM-XRC-5TZ
**
**   with
**
**   o ADE-XRC
**     SlimBridge PCI-express module
**
** (C) Copyright 2008 Alpha Data
**
**
** NOTES ON OPERATION
**
** This application is designed to demonstrate the use of the
** SlimBridge PCI-e DMA ontroller
*/

#include <windows.h>
#include <stdio.h>

#include <admxrc2.h>
#include <adexrc.h>

#include "..\common\common.h"


/*
** Test Application Configuration Items
**
** Change these settings to alter behaviour of the test application.
*/

// #define PERFORM_CHECK       (1)         /* 1 => Check data received      */

// #define PERFORMANCE_TEST    (1)         /* 1 => Run performance test     */

#define INTERACTIVE         (1)         /* 1 => Press enter to display counters */

#define DMA_BUFFER_SIZE     (8*1024*1024)

#define NUM_PKTS_IN_FLIGHT  (1)         /* Number of packets in flight over PCIe bus */

#define PCIE_BUFFER_SIZE_TX    (0x20)      /* PCIe Buffer Size - 0x20 for 128 bytes and 0x10 for 64 bytes */
#define PCIE_BUFFER_SIZE_RX    (0x200)      /* PCIe Buffer Size - 0x20 for 128 bytes and 0x10 for 64 bytes */

#define DRIVER_DELAY        10         /* Delay to ensure ADM-XRC and ADE-XRC drivers are in sync */

/*
** The next two constants shouldn't be changed unless code is
** added to enforce mutual exclusion when setting the transfer
** counters
*/
#define NUM_BLOWER_THREAD   (1)
#define NUM_SUCKER_THREAD   (1)         /* Must be 1 if PERFORM_CHECK    */

static ADMXRC2_HANDLE       card = NULL;
static ADEXRC_HANDLE        card1 = NULL;

static HANDLE               txReadySema = NULL;
static HANDLE               rxReadySema = NULL;

static BOOLEAN              terminate;
static BOOLEAN              terminateSucker;

static HANDLE               sucker[NUM_SUCKER_THREAD], blower[NUM_BLOWER_THREAD];
static DWORD                suckerId[NUM_SUCKER_THREAD], blowerId[NUM_BLOWER_THREAD];
static double               got[NUM_SUCKER_THREAD], sent[NUM_BLOWER_THREAD];
static LARGE_INTEGER        got_time[NUM_SUCKER_THREAD], sent_time[NUM_BLOWER_THREAD];
static LARGE_INTEGER        perffreq;

static DWORD                bufSize;

static ADEXRC_DMADESC       sendbufHandle;
static BOOLEAN              sendbufValid = FALSE;
static DWORD*               sendbuf = NULL;
static ADEXRC_DMADESC       recvbufHandle;
static BOOLEAN              recvbufValid = FALSE;
static DWORD*               recvbuf = NULL;

static volatile DWORD*      fpgaSpace;


float GetTimeInMilliSeconds(LARGE_INTEGER time)
{
	return ((float) (time.QuadPart*1000i64)) / ((float) perffreq.QuadPart);
}

DWORD WINAPI blowerThread(PVOID arg)
{
    ADEXRC_STATUS       status;
    int                 ii = (int) arg;
	int i;
    DWORD               j;
    HANDLE              event;
    DWORD*              buf = (DWORD*) (sendbuf);
    DWORD               data;
    LARGE_INTEGER       start, now, delta;

    event = CreateEvent(NULL, TRUE, FALSE, NULL);

    /* Set up the data */
    for (j = 0; j < (bufSize >> 2); j++) {
        buf[j] = j;
    }

    // Set up FIFO read in FPGA application
    data = bufSize; // Set transfer size     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,0x0,&data,4);
    data = 16;  // Set burst size       
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,12*4,&data,4);
    data = 16;  // Set row size    
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,13*4,&data,4);
    data = 65536;  // Set nrows     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,14*4,&data,4);
    data = 16;  // Set end row skip   
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,15*4,&data,4);
    
    while (!terminate) {
        if (WaitForSingleObject(txReadySema, INFINITE) != WAIT_OBJECT_0) {
            printf("Blower thread %ld: failed to wait on semaphore\n", (unsigned long) i);
            break;
        }
Sleep(DRIVER_DELAY);
/*
status = ADEXRC_ReadReg(card1, 0x11C, &data );
    printf("Before Delay PCIe Register 0x%lX = 0x%lX)\n", 0x11C, data ); 
    printf("Debug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");
Sleep(DRIVER_DELAY);

status = ADEXRC_ReadReg(card1, 0x11C, &data );
    printf("After Delay PCIe Register 0x%lX = 0x%lX)\n", 0x11C, data ); 
	    printf("Debug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");//*/
#if PERFORMANCE_TEST
        for( i = 0; i < 1000; i++ )
#endif
		{
            /*
             * Setting up OCP DMA engine to operate in demand
             * mode, reading the data from the PCI-E FIFO
             */
            data = 0;  // Start Address : Start OCP DMA engine
            ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,16*4,&data,4);

            /* Make OCP DMA engine has started */
            //Sleep(DRIVER_DELAY);

			QueryPerformanceCounter(&start);

            /*
             * Start the DMA transfer over PCIe
             */
            status = ADEXRC_DoDMA( card1,
                                   sendbufHandle,
                                   ADEXRC_PCITOLOCAL,
                                   0,
                                   bufSize,
                                   0,
                                   0,
                                   event);
            if (status != ADEXRC_SUCCESS) {
                printf("Blower thread %ld: failed to do DMA: %s\n",
                       (unsigned long) i, ADEXRC_GetStatusString(status));
                break;
            }
        
			QueryPerformanceCounter(&now);
/*
    status = ADEXRC_ReadReg(card1, 0x11C, &data );
    printf("After DMA PCIe Register 0x%lX = 0x%lX)\n", 0x11C, data );
	    printf("Debug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");//*/
            sent[ii] += bufSize;
            delta.QuadPart = now.QuadPart - start.QuadPart;
            sent_time[ii].QuadPart += delta.QuadPart;
        }

        ReleaseSemaphore(rxReadySema, 1, NULL);
    }

    CloseHandle(event);

    return 0;
}

DWORD WINAPI suckerThread(PVOID arg)
{
    ADMXRC2_STATUS      status;
    int                 ii = (int) arg;
	int i;
#if PERFORM_CHECK
    DWORD               j;
    DWORD               act;
	DWORD               exp;
#endif
    HANDLE              event;
    DWORD*              buf = (DWORD*) (recvbuf);
    DWORD               n = 0;
    DWORD               data;
    LARGE_INTEGER       start, now, delta;

    event = CreateEvent(NULL, TRUE, FALSE, NULL);

    // Set up FIFO write in FPGA application
    data = bufSize; // Set transfer size     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,0x4,&data,4);
    data = 16;  // Set burst size       
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,17*4,&data,4);
    data = 16;  // Set row size    
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,18*4,&data,4);
    data = 65536;  // Set nrows     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,19*4,&data,4);
    data = 16;  // Set end row skip   
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,20*4,&data,4);
    
    while (1) {
        if (WaitForSingleObject(rxReadySema, INFINITE) != WAIT_OBJECT_0) {
            printf("Sucker thread %ld: failed to wait on semaphore\n", (unsigned long) i);
            break;
        }

        if (terminateSucker) {
            break;
        }
 /*      
		status = ADEXRC_ReadReg(card1, 0x11C, &data );
    printf("Before Read DMA PCIe Register 0x%lX = 0x%lX)\n", 0x11C, data ); 
	    printf("Debug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");//*/
#if PERFORMANCE_TEST
        for( i = 0; i < 1000; i++ )
#endif
     {
            data = 0;  // Start Address : Start OCP DMA engine
            ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,21*4,&data,4);
            
            /* Make OCP DMA engine has started */
            //Sleep(DRIVER_DELAY);
            
			QueryPerformanceCounter(&start);
            /*
             * Start the DMA transfer over PCIe
             */
            status = ADEXRC_DoDMA( card1,
                                   recvbufHandle,
                                   ADEXRC_LOCALTOPCI,
                                   0,
                                   bufSize,
                                   0,
                                   0,
                                   event);
            if (status != ADEXRC_SUCCESS) {
                printf("Sucker thread %ld: failed to do DMA: %s\n",
                       (unsigned long) i, ADEXRC_GetStatusString(status));
                break;
            }
			QueryPerformanceCounter(&now);
/*			    
    printf("After Read DMADebug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");//*/
//Sleep(DRIVER_DELAY);
//Sleep(DRIVER_DELAY);
		//	   status = ADEXRC_ReadReg(card1, 0x11C, &data );
		//	   printf("After Read DMA PCIe Register 0x%lX = 0x%lX)\n", 0x11C, data );  
            got[ii] += bufSize;
            delta.QuadPart = now.QuadPart - start.QuadPart;
            got_time[ii].QuadPart += delta.QuadPart;

#if PERFORM_CHECK
            /* Check the data */
            for (j = 0; j < (bufSize >> 2); j++) {
                act = buf[j];
				if (j==0) 
				  exp=act;
                if ((act&0xFFFF) != (exp&0xFFFF)) {
                    printf("Sucker thread %ld: error at block %ld index %ld, exp/act=0x%8.8lx/0x%8.8lx\n",
                           (unsigned long) ii, (unsigned long) n, (unsigned long) j,
                           (unsigned long) exp, (unsigned long) act);
                }
				exp++;
				j++;
				act = buf[j];
				if (act != 0xDEADF1F0) {
                    printf("Sucker thread %ld: error at block %ld index %ld, exp/act=0x%8.8lx/0x%8.8lx\n",
                           (unsigned long) ii, (unsigned long) n, (unsigned long) j,
                           (unsigned long) 0xDEADF1F0, (unsigned long) act);
                }
            }
#endif
        }
        
        n++;

        ReleaseSemaphore(txReadySema, 1, NULL);
    }

    CloseHandle(event);

    return 0;
}

int main(int argc, char *argv[])
{
    int                 ret = 0;
    ADMXRC2_STATUS      status;
    ADMXRC2_CARD_INFO   cardInfo;
    ADEXRC_STATUS       status1;
    ADEXRC_CARD_INFO    cardInfo1;
    int                 i;
    HANDLE              event = NULL;
    Arguments           arguments;
    Option              options[4];
    float               clock_freq;
    DWORD               data;
    
    options[0].key = "card";
    options[0].type = Option_uint;
    options[0].def.uintVal = 0;
    options[0].meaning = "ID of card to use";
    options[1].key = "index";
    options[1].type = Option_uint;
    options[1].def.uintVal = 0;
    options[1].meaning = "index of card to use";
    options[2].key = "size";
    options[2].type = Option_uint;
    options[2].def.uintVal = DMA_BUFFER_SIZE;
    options[2].meaning = "send/receive buffer size (bytes)";
    options[3].key = "lclk";
    options[3].type = Option_float;
    options[3].def.floatVal = 33.0;
    options[3].meaning = "local bus clock frequency (MHz)";
    arguments.nOption = 4;
    arguments.option = options;
    arguments.nParamType = 0;
    arguments.minNParam = 0;

    parse_command_line(
        "dma",
        argc,
        argv,
        &arguments);

    if (options[0].specified && options[1].specified) {
        printf("Do not specify both /card and /index\n");
        ret = -1;
        goto done;
    }

    if (!options[1].specified) {
        ADMXRC2_CARDID cardID = options[0].value.uintVal;
        ADEXRC_CARDID card1ID;

        status = ADMXRC2_OpenCard(cardID, &card);
        if (status != ADMXRC2_SUCCESS) {
            printf("Failed to open PCI-X card with ID %ld: %s\n",
                (unsigned long) cardID, ADMXRC2_GetStatusString(status));
            ret = -1;
            goto done;
        }
        
        ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, 80.0e6, NULL);
        ADMXRC2_SetClockRate(card, 1, 133.0e6, NULL);
        data = 1;
        ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,ADMXRC2_IOADAPTER,0x28,&data,4);
        Sleep(1);
        data = 0;
        ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,ADMXRC2_IOADAPTER,0x28,&data,4);
        Sleep(1);

        /*
         * Set the bridge ID for the PCIe card. We can either use this value
         * to open a specific card, or 0 to open the first one we find.
         */
        data = 0x12345672;
        ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,0x10,&data,4);

        card1ID = 0x12345672;   /* Or 0 to open the first/only card */
        status = ADEXRC_OpenCard(&card1, card1ID);
        if (status != ADEXRC_SUCCESS) {
            printf("Failed to open PCIe card with ID %ld: %s\n",
                (unsigned long) card1ID, ADEXRC_GetStatusString(status));
            ret = -1;
            goto done;
        }
    } else {
        unsigned int index = options[1].value.uintVal;

        status = ADMXRC2_OpenCardByIndex(index, &card);
        if (status != ADMXRC2_SUCCESS) {
            printf("Failed to open PCI-X card with index %ld: %s\n",
                (unsigned long) index, ADMXRC2_GetStatusString(status));
            ret = -1;
            goto done;
        }

        status1 = ADEXRC_OpenCard(&card1, index);
        if (status1 != ADEXRC_SUCCESS) {
            printf("Failed to open PCIe card with index %ld: %s\n",
                (unsigned long) index, ADEXRC_GetStatusString(status1));
            ret = -1;
            goto done;
        }
    }

    /* Allocate memory for the DMA data buffers */
    bufSize = options[2].value.uintVal;
    if( bufSize > DMA_BUFFER_SIZE ){
        printf("Buffer size must be less than or equal to %d\n", DMA_BUFFER_SIZE);
        ret = -1;
        goto done;
    }
    /* Get a DMA descriptor for the buffer used for DMA transfers to the card */
    sendbuf = malloc(DMA_BUFFER_SIZE);
    if (sendbuf == NULL) {
        printf("Failed to allocate buffers for threads\n");
        ret = -1;
        goto done;
    }

    /* Get a DMA descriptor for the buffer used for DMA transfers from the card */
    recvbuf = malloc(DMA_BUFFER_SIZE);
    if (recvbuf == NULL) {
        printf("Failed to allocate receive buffer for threads\n");
        ret = -1;
        goto done;
    }
    
    clock_freq = options[3].value.floatVal * 1.0e6f;

    /*
    ** Create a manual reset event for use when configuring the FPGA
    ** using DMA.
    */
    event = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (event == NULL) {
        printf("Failed to create Win32 Event\n");
        ret = -1;
        goto done;
    }

    /* Get card information, for device fitted */
    status = ADMXRC2_GetCardInfo(card, &cardInfo);
    if (status != ADMXRC2_SUCCESS) {
        printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
        ret = -1;
        goto done;
    }

    /* Get card information, for PCIe device fitted */
    status1 = ADEXRC_GetCardInfo(card1, &cardInfo1);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to get card info: %s\n", ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }

    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
        break;

    default:
        printf("This example must be run on one of the following cards:\n\n");
        printf("\tADM-XRC-4FX\n");
        printf("\tADPE-XRC-4FX\n");
        printf("\tADM-XRC-5LX\n");
        printf("\tADM-XRC-5T1\n");
        printf("\tADM-XRC-5T2\n");
        printf("\tADM-XRC-5TZ\n");
        ret = -1;
        goto done;
    }
    
    /*
    ** Semaphores are required for each thread to notify the other thread
    ** that the other thread can proceed with its DMA.
    */
    txReadySema = CreateSemaphore(NULL, 1, 1024, NULL);
    if (txReadySema == NULL) {
        printf("Failed to create Win32 Semaphore\n");
        ret = -1;
        goto done;
    }
    rxReadySema = CreateSemaphore(NULL, 0, 1024, NULL);
    if (rxReadySema == NULL) {
        printf("Failed to create Win32 Semaphore\n");
        ret = -1;
        goto done;
    }

    /* Set the number of buffers in flight on PCIe */
    printf("Writing 0x%X to PCIe Register 0x4C\n", NUM_PKTS_IN_FLIGHT ); 
    data = NUM_PKTS_IN_FLIGHT;
    status1 = ADEXRC_WriteReg(card1, 0x4C, data );
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to write register: %s\n",
               ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }

    /* Set the PCIe buffer size */
    printf("Writing 0x%X to PCIe Register 0x0C\n", PCIE_BUFFER_SIZE_TX ); 
    data = PCIE_BUFFER_SIZE_TX;
    status1 = ADEXRC_WriteReg(card1, 0x0C, data );
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to write register: %s\n",
               ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }
    printf("Writing 0x%X to PCIe Register 0x20\n", PCIE_BUFFER_SIZE_RX );
    data = PCIE_BUFFER_SIZE_RX;
    status1 = ADEXRC_WriteReg(card1, 0x20, data );
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to write register: %s\n",
               ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }

    status1 = ADEXRC_SetupDMA(card1, sendbuf, DMA_BUFFER_SIZE, 0, &sendbufHandle);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to set up send DMA buffer: %s\n", ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }
    sendbufValid = TRUE;

    status1 = ADEXRC_SetupDMA(card1, recvbuf, DMA_BUFFER_SIZE, 0, &recvbufHandle);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to set up receive DMA buffer: %s\n", ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }
    recvbufValid = TRUE;

    /*
    ** Start up the blower and sucker DMA threads.
    */
    terminate = FALSE;
    terminateSucker = FALSE;
    printf("Starting %ld blower thread(s)...\n", (unsigned long) NUM_BLOWER_THREAD);
    for (i = 0; i < NUM_BLOWER_THREAD; i++) {
        sent[i] = 0.0;
        blower[i] = CreateThread(
            NULL,
            bufSize + 16384,
            blowerThread,
            (PVOID) i,
            0,
            &blowerId[i]);
    }
    printf("Starting %d sucker thread(s)...\n", (unsigned long) NUM_SUCKER_THREAD);
    for (i = 0; i < NUM_SUCKER_THREAD; i++) {
        got[i] = 0.0;
        sucker[i] = CreateThread(
            NULL,
            bufSize + 16384,
            suckerThread,
            (PVOID) i,
            0,
            &suckerId[i]);
    }
    
    
    Sleep(200);
    
    /* Display average transfer rates for ever or until user enters 'q' */
	if (QueryPerformanceFrequency(&perffreq)) {

	}

	for (i = 0; i < NUM_BLOWER_THREAD; i++) {
		sent_time[i].QuadPart = 0;
        got_time[i].QuadPart = 0;
    }

    while (1) {
        int     c;
        double  s, g, t;
		LARGE_INTEGER st, gt, tt;

        s = 0.0; g = 0.0;
        st.QuadPart = 0; gt.QuadPart = 0;
        for (i = 0; i < NUM_BLOWER_THREAD; i++) {
            s += sent[i];
            st.QuadPart += sent_time[i].QuadPart;
        }
        for (i = 0; i < NUM_SUCKER_THREAD; i++) {
            g += got[i];
            gt.QuadPart += got_time[i].QuadPart;
        }
        s = s / 1048576.0f;
        g = g / 1048576.0f;
        t = s + g;
        tt.QuadPart = st.QuadPart + gt.QuadPart;

        printf("Sent %.2f MiB, received %.2f MiB, total %.2f MiB\n",
            s, g, t);
        if (st.QuadPart != 0 && gt.QuadPart != 0) {
            printf("Send rate %.2f MiB/s, receive rate %.2f MiB/s, total rate %.2f MiB/s\n",
                s * 1000.0 / GetTimeInMilliSeconds(st),
                g * 1000.0 / GetTimeInMilliSeconds(gt),
                t * 1000.0 / GetTimeInMilliSeconds(tt));
        }

#if INTERACTIVE
        printf("Enter 'q' to exit, anything else to view rates:\n");
        c = getchar();
        if (c == 'q') {
            break;
        }
        if (c == 'd') {
            printf("Debug Register Status: ");
            for (i=0;i<9;i++)
            {
                ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
                printf("%x ",data);
            }
            printf("\n");
      
            printf("PCIe FIFO\n");
            for (i=0;i<1024;i++)
            {
                char ch;
                ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,16*4+0x7F800,&data,4);
                printf("%x ",data);
                ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,17*4+0x7F800,&data,4);
                printf("%x ",data);
                ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,255*4+0x7F800,&data,4);
                if (i%4==3)
                    printf("\n");
                if (i%64==63)
                    ch=getchar();
                if (ch=='q')
                    break;
            }
        }
        Sleep(100);
#else
        Sleep(500);
#endif
    }

    /*
    ** Terminate the threads cleanly.
    **
    ** First terminate all blower DMA threads. We wait for 100ms
    ** for each thread to terminate. If a thread doesn't terminate
    ** within 100ms, assume something has gone wrong (shouldn't
    ** happen) and attempt to terminate it by brute force via
    ** TerminateThread().
    **
    ** Next, terminate the sucker DMA thread. We increment the
    ** semaphore once in case the sucker thread is currently
    ** waiting on the semaphore. If it doesn't terminate within
    ** 100ms, we use TerminateThread() as with the blower threads.
    */
    printf("Terminating threads...\n");
    terminate = TRUE;
    ReleaseSemaphore(txReadySema, 1, NULL);
    for (i = 0; i < NUM_BLOWER_THREAD; i++) {
        if (WaitForSingleObject(blower[i], 100) != WAIT_OBJECT_0) {
            if (!TerminateThread(blower[i], 0)) {
                printf("Failed to terminate blower thread %ld\n", (unsigned long) i);
            }
        }
    }
    terminateSucker = TRUE;
    ReleaseSemaphore(rxReadySema, 1, NULL);
    for (i = 0; i < NUM_SUCKER_THREAD; i++) {
        if (WaitForSingleObject(sucker[i], 100) != WAIT_OBJECT_0) {
            if (!TerminateThread(sucker[i], 0)) {
                printf("Failed to terminate sucker thread %ld\n", (unsigned long) i);
            }
        }
    }

done:
    /* Free the DMA descriptors */
    if (sendbufValid) {
        status1 = ADEXRC_UnsetupDMA(card1, sendbufHandle);
        if (status1 != ADEXRC_SUCCESS) {
            printf("Failed to delete DMA send buffer: %s\n",
                   ADEXRC_GetStatusString(status1));
        }
    }
    if (recvbufValid) {
        status = ADEXRC_UnsetupDMA(card1, recvbufHandle);
        if (status1 != ADEXRC_SUCCESS) {
            printf("Failed to delete DMA receive buffer: %s\n",
                   ADEXRC_GetStatusString(status1));
        }
    }

    /* Free the Win32 Semaphore */
    if (txReadySema != NULL) {
        CloseHandle(txReadySema);
    }
    if (rxReadySema != NULL) {
        CloseHandle(rxReadySema);
    }

    /* Free the Win32 Event */
    if (event != NULL) {
        CloseHandle(event);
    }

    /* Close the device */
    if (card != NULL) {
        ADMXRC2_CloseCard(card);
    }
    if (card1 != NULL) {
        ADEXRC_CloseCard(card1);
    }
    
    /* Free the memory used for the DMA data buffers */
    if (sendbuf != NULL) {
        free(sendbuf);
    }
    if (recvbuf != NULL) {
        free(recvbuf);
    }

    return ret;
}
