/*
** simpledma.c - trivial example to demonstrate DMA transfers
**               to and from FPGA, using demand mode DMA.
**
** Data is transferred using demand mode DMA to the FPGA, stored
** in a FIFO and transferred out again also using demand mode DMA.
** Also demonstrated is configuring the FPGA using DMA.
**
** Example for:
**
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2
**   o ADM-XRC-5TZ
**
**   with
**
**   o ADE-XRC
**     SlimBridge PCI-express module
**
** (C) Copyright 2008 Alpha Data
**
**
** NOTES ON OPERATION
**
** This application is designed to demonstrate the use of the
** SlimBridge PCI-e DMA ontroller
*/

#include <windows.h>
#include <stdio.h>

#include <admxrc2.h>
#include <adexrc.h>

#include "..\common\common.h"

/*
** Test Application Configuration Items
**
** Change these settings to alter behaviour of the test application.
*/

// #define DEBUG                           /* Uncomment to dump PCIe registers    */

#define PERFORM_CHECK       (1)         /* 1 => Check data received      */

#define DMA_BUFFER_SIZE     (1024*1024)

#define NUM_PKTS_IN_FLIGHT  (1)

#define DRIVER_DELAY        250

static ADMXRC2_HANDLE       card = NULL;
static ADEXRC_HANDLE        card1 = NULL;

static ULONG                bufSize;

static ADEXRC_DMADESC       sendbufHandle;
static BOOLEAN              sendbufValid = FALSE;
static DWORD*               sendbuf = NULL;
static ADEXRC_DMADESC       recvbufHandle;
static BOOLEAN              recvbufValid = FALSE;
static DWORD*               recvbuf = NULL;

static volatile DWORD*      fpgaSpace;



int main(int argc, char *argv[])
{
    int                 ret = 0;
    ADMXRC2_STATUS      status;
    ADMXRC2_CARD_INFO   cardInfo;
    ADEXRC_STATUS       status1;
    ADEXRC_CARD_INFO    cardInfo1;
    DWORD               i;
    HANDLE              event = NULL;
    Arguments           arguments;
    Option              options[3];
    DWORD               data;
    int                 errs;
    
    options[0].key = "card";
    options[0].type = Option_uint;
    options[0].def.uintVal = 0;
    options[0].meaning = "ID of card to use";
    options[1].key = "index";
    options[1].type = Option_uint;
    options[1].def.uintVal = 0;
    options[1].meaning = "index of card to use";
    options[2].key = "size";
    options[2].type = Option_uint;
    options[2].def.uintVal = 64*1024;
    options[2].meaning = "send/receive buffer size (bytes)";
    arguments.nOption = 3;
    arguments.option = options;
    arguments.nParamType = 0;
    arguments.minNParam = 0;

    parse_command_line(
        "dma",
        argc,
        argv,
        &arguments);

    if (options[0].specified && options[1].specified) {
        printf("Do not specify both /card and /index\n");
        ret = -1;
        goto done;
    }

    if (!options[1].specified) {
        ADMXRC2_CARDID cardID = options[0].value.uintVal;
        ADEXRC_CARDID card1ID;

        status = ADMXRC2_OpenCard(cardID, &card);
        if (status != ADMXRC2_SUCCESS) {
            printf("Failed to open PCI-X card with ID %ld: %s\n",
                (unsigned long) cardID, ADMXRC2_GetStatusString(status));
            ret = -1;
            goto done;
        }
        
        /* Set up card */
        ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, 66.0e6, NULL);
        ADMXRC2_SetClockRate(card, 1, 133.0e6, NULL);
        data = 1;
        ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,ADMXRC2_IOADAPTER,0x28,&data,4);
        Sleep(1);
        data = 0;
        ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,ADMXRC2_IOADAPTER,0x28,&data,4);
        Sleep(1);

        /*
         * Set the bridge ID for the PCIe card. We can either use this value
         * to open a specific card, or 0 to open the first one we find.
         */
		Sleep(350);
        data = 0x12345678;
        ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,0x10,&data,4);

        card1ID = 0x12345678;   /* Or 0 to open the first/only card */
        status = ADEXRC_OpenCard(&card1, card1ID);
        if (status != ADEXRC_SUCCESS) {
            printf("Failed to open PCIe card with ID %ld: %s\n",
                (unsigned long) card1ID, ADEXRC_GetStatusString(status));
            ret = -1;
            goto done;
        }
    } else {
        unsigned int index = options[1].value.uintVal;

        status = ADMXRC2_OpenCardByIndex(index, &card);
        if (status != ADMXRC2_SUCCESS) {
            printf("Failed to open PCI-X card with index %ld: %s\n",
                (unsigned long) index, ADMXRC2_GetStatusString(status));
            ret = -1;
            goto done;
        }

        status1 = ADEXRC_OpenCard(&card1, index);
        if (status1 != ADEXRC_SUCCESS) {
            printf("Failed to open PCIe card with index %ld: %s\n",
                (unsigned long) index, ADEXRC_GetStatusString(status1));
            ret = -1;
            goto done;
        }
    }

    /* Get card information, for device fitted */
    status = ADMXRC2_GetCardInfo(card, &cardInfo);
    if (status != ADMXRC2_SUCCESS) {
        printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
        ret = -1;
        goto done;
    }

    /* Get card information, for PCIe device fitted */
    status1 = ADEXRC_GetCardInfo(card1, &cardInfo1);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to get card info: %s\n", ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }

    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
        break;

    default:
        printf("This example must be run on one of the following cards:\n\n");
        printf("\tADM-XRC-4FX\n");
        printf("\tADPE-XRC-4FX\n");
        printf("\tADM-XRC-5LX\n");
        printf("\tADM-XRC-5T1\n");
        printf("\tADM-XRC-5T2\n");
        printf("\tADM-XRC-5TZ\n");
        ret = -1;
        goto done;
    }
    
    /* Get a DMA descriptor for the buffer used for DMA transfers to the card */
    sendbuf = malloc(DMA_BUFFER_SIZE);
    /* Fill buffer with incrementing pattern */
    for (i=0;i<DMA_BUFFER_SIZE/4;i++) {
        sendbuf[i] = i;
    }
    status1 = ADEXRC_SetupDMA(card1, sendbuf, DMA_BUFFER_SIZE, 0, &sendbufHandle);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to set up send DMA buffer: %s\n", ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }
    sendbufValid = TRUE;

    /* Get a DMA descriptor for the buffer used for DMA transfers from the card */
    recvbuf = malloc(DMA_BUFFER_SIZE);
    /* Fill buffer with zeroes */
    for (i=0;i<DMA_BUFFER_SIZE/4;i++) {
        recvbuf[i] = 0;
    }
    status1 = ADEXRC_SetupDMA(card1, recvbuf, DMA_BUFFER_SIZE, 0, &recvbufHandle);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to set up receive DMA buffer: %s\n", ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }
    recvbufValid = TRUE;

    /* Set the number of buffers in flight */
    printf("Writing 0x%X to PCIe Register 0x4C\n", NUM_PKTS_IN_FLIGHT ); 
    data = NUM_PKTS_IN_FLIGHT;
    status1 = ADEXRC_WriteReg(card1, 0x4C, data );
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to write register: %s\n",
               ADEXRC_GetStatusString(status1));
        ret = -1;
        goto done;
    }
    ADEXRC_WriteReg(card1, 0x0C, 0x20 );
    ADEXRC_WriteReg(card1, 0x20, 0x200 );

    // Set up 1MB FIFO read in FPGA application
    data = DMA_BUFFER_SIZE; // Set transfer size     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,0x0,&data,4);
    data = 32;  // Set burst size       
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,12*4,&data,4);
    data = 32;  // Set row size    
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,13*4,&data,4);
    data = 65536;  // Set nrows     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,14*4,&data,4);
    data = 32;  // Set end row skip   
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,15*4,&data,4);      

    /*
     * Setting up OCP DMA engine to operate in demand
     * mode, reading the data from the PCI-E FIFO
     */
    data = 0;  // Start Address : Start OCP DMA engine
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,16*4,&data,4);
    
    /* Make sure OCP DMA engine has started */
    //Sleep(DRIVER_DELAY);
	goto write;
    printf("\nStart PCIe DMA Read\n");

    /*
     * Start the DMA transfer over PCIe
     */
    status1 = ADEXRC_DoDMA( card1,
                            sendbufHandle,
                            ADEXRC_PCITOLOCAL,
                            0,
                            DMA_BUFFER_SIZE,
                            0,
                            0,
                            event);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to start DMA Read operation: %s\n", 
               ADEXRC_GetStatusString(status1));
        goto done;
    }

//#ifdef DEBUG 
	printf("OCP Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x88,&data,4);
        printf("%8.8x %8.8x \n",i*4+0x88,data);
    }
    printf("\n");

    printf("Debug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");
      /*
    printf("PCIe FIFO\n");
    for (i=0;i<1024;i++)
    {
        char c;
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,16*4+0x7F800,&data,4);
        printf("%x ",data);
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,17*4+0x7F800,&data,4);
        printf("%x ",data);
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,255*4+0x7F800,&data,4);
        if (i%4==3)
            printf("\n");
        if (i%64==63)
            c=getchar();
        if (c=='q')
            break;
    }*/
//#endif
    
    // Clear buffer
    for (i=0;i<DMA_BUFFER_SIZE/4;i++) {
        recvbuf[i] = 0xFFFFFFFF;
    }
write:   
    // Set up 1MB FIFO write in FPGA application
    data = DMA_BUFFER_SIZE; // Set transfer size     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,0x4,&data,4);
    data = 32;  // Set burst size       
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,17*4,&data,4);
    data = 32;  // Set row size    
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,18*4,&data,4);
    data = 65536;  // Set nrows     
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,19*4,&data,4);
    data = 32;  // Set end row skip   
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,20*4,&data,4);
    data = 0;  // Start Address : Start OCP DMA engine
    ADMXRC2_Write(card,ADMXRC2_IOWIDTH_32,0,21*4,&data,4);

    /*
     * OCP Engine now reading data from FPGA local memory
     * to PCI-E FIFO so it can be transferred back over
     * PCI-E bus to the CPU memory
     */
            
    //Sleep(DRIVER_DELAY);

    printf("\nStart PCIe DMA Write\n");

    /*
     * Start the DMA transfer over PCIe
     */
    status1 = ADEXRC_DoDMA( card1,
                            recvbufHandle,
                            ADEXRC_LOCALTOPCI,
                            0,
                            DMA_BUFFER_SIZE,
                            0,
                            0,
                            event);
    if (status1 != ADEXRC_SUCCESS) {
        printf("Failed to start DMA Write operation: %s\n", 
               ADEXRC_GetStatusString(status1));
        goto done;
    }
    
    errs = 0;
    // Check Data
    for (i=0;i<DMA_BUFFER_SIZE/4;i++)
    {
        if (recvbuf[i] != i)
        {
            errs++;
            if (errs<32)
                printf("Error at %d : expected %x, read %x\n",i,i,recvbuf[i]);
            if (errs==32)
                printf("More errors...\n"); 
        }
    }
    if( !errs ) {
        printf("DMA Read/Write Test passed!\n");
    }
    
//#ifdef DEBUG
    printf("OCP Register Status: \n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x88,&data,4);
        printf("%8.8x %8.8x \n",i*4+0x88,data);
    }
    printf("\n");

    printf("Debug Register Status:\n");
    for (i=0;i<10;i++)
    {
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,i*4+0x7F800,&data,4);
        printf("%8.8x \n",data);
    }
    printf("\n");
      
   /* printf("PCIe FIFO\n");
    for (i=0;i<1024;i++)
    {
        char c;
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,16*4+0x7F800,&data,4);
        printf("%x ",data);
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,17*4+0x7F800,&data,4);
        printf("%x ",data);
        ADMXRC2_Read(card,ADMXRC2_IOWIDTH_32,0,255*4+0x7F800,&data,4);
        if (i%4==3)
            printf("\n");
        if (i%64==63)
            c=getchar();
        if (c=='q')
            break;
    }*/
//#endif
    
    Sleep(200);
    

done:
    /* Free the DMA descriptors */
    if (sendbufValid) {
        status1 = ADEXRC_UnsetupDMA(card1, sendbufHandle);
        if (status1 != ADEXRC_SUCCESS) {
            printf("Failed to delete DMA send buffer: %s\n",
                   ADEXRC_GetStatusString(status1));
        }
    }
    if (recvbufValid) {
        status = ADEXRC_UnsetupDMA(card1, recvbufHandle);
        if (status1 != ADEXRC_SUCCESS) {
            printf("Failed to delete DMA receive buffer: %s\n",
                   ADEXRC_GetStatusString(status1));
        }
    }

    /* Free the Win32 Event */
    if (event != NULL) {
        CloseHandle(event);
    }

    /* Close the device */
    if (card != NULL) {
        ADMXRC2_CloseCard(card);
    }
    if (card1 != NULL) {
        ADEXRC_CloseCard(card1);
    }
    
    /* Free the memory used for the DMA data buffers */
    if (sendbuf != NULL) {
        free(sendbuf);
    }
    if (recvbuf != NULL) {
        free(recvbuf);
    }

    return ret;
}
